import 'dart:async';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:get/get.dart';
import 'package:permission_handler/permission_handler.dart';
import 'package:rdc_flutter_sdk/rabbit_data.dart';
import 'package:rdc_flutter_sdk/src/controllers/condition_controller.dart';
import 'package:realm/realm.dart';
import 'package:retry/retry.dart';
import 'package:rdc_flutter_sdk/src/constants/app_constants.dart' as con;
import 'package:rdc_flutter_sdk/src/controllers/device_info_controller.dart';
import 'package:rdc_flutter_sdk/src/controllers/location_controller.dart';
import 'package:rdc_flutter_sdk/src/controllers/rabbit_config_controller.dart';
import 'package:rdc_flutter_sdk/src/controllers/user_info_controller.dart';
import 'package:rdc_flutter_sdk/src/controllers/version_controller.dart';
import 'package:rdc_flutter_sdk/src/network/api_connector.dart';
import 'package:rdc_flutter_sdk/src/repositories/event_repo.dart';
import 'package:rxdart/subjects.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:pausable_timer/pausable_timer.dart';

import '../util/date_time_helper.dart';

class EventController {
  /* ------------ service ------------*/
  late ApiConnector _api;
  /* ------------ service ------------*/

  /* ------------ repo ------------*/
  late EventRepo _eventRepo;
  /* ------------ repo ------------*/

  /* ------------ controller ------------*/
  final RabbitConfigController _config = Get.find(tag: 'rdc_flutter_sdk');
  final ConditionController _conditionController =
      Get.find(tag: 'rdc_flutter_sdk');
  final DeviceInfoController _deviceInfoController =
      Get.find(tag: 'rdc_flutter_sdk');
  final UserInfoController _userInfoController =
      Get.find(tag: 'rdc_flutter_sdk');
  final LocationController _locationController =
      Get.find(tag: 'rdc_flutter_sdk');
  final VersionController _versionController = Get.find(tag: 'rdc_flutter_sdk');
  /* ------------ controller ------------*/

  /* ------------ variable ------------*/
  final RabbitRequest rabbitRequest = RabbitRequest();
  var session = RabbitSession.empty;
  var _sessionId = '';
  DateTime _sessionStart = DateTime.now();
  int? loopTime;
  PausableTimer? sessionTask;
  PausableTimer? resumeTask;
  PausableTimer? rotationTask;

  var isLoading = false;
  /* ------------ variable ------------*/

  /* ------------ stream ------------*/
  var changeDB = BehaviorSubject.seeded('');
  var apiResult = BehaviorSubject.seeded('');
  /* ------------ stream ------------*/

  Future<EventController> init() async {
    _eventRepo = await EventRepo().init();
    _api = await ApiConnector().init();
    _initRequest();
    await checkOpenApp();
    await rotationLoop();
    await _reserve();
    // await _clearData();
    return this;
  }

  Future<void> _reserve() async {
    if (_config.dbType == DBType.hive.name) return;

    final sharedPreference = await SharedPreferences.getInstance();

    if (sharedPreference.containsKey(con.realmReserved) &&
        _config.dbType == DBType.realm.name) return;

    // 1000+ char
    var valueStr =
        '''xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx
    xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx
    xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx
    xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx
    xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx
    xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx
    xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx
    xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx
    xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx
    xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx xxxxxxxxxx''';

    var list = List<RabbitRequest>.generate(100, (index) {
      var data = RabbitRequest()
        ..id = index + 1
        ..sdkVersion = valueStr
        ..globalId = valueStr
        ..osType = valueStr
        ..osVer = valueStr
        ..display = valueStr
        ..maid = valueStr
        ..device = valueStr
        ..model = valueStr
        ..brandName = valueStr
        ..board = valueStr
        ..memberId = valueStr
        ..ssoId = valueStr
        ..sessionId = valueStr
        ..sessionEvent = valueStr
        ..sdkVersion = valueStr
        ..eventName = valueStr
        ..eventValue = valueStr
        ..osVer = valueStr
        ..localeCountry = valueStr
        ..localeLang = valueStr
        ..latitude = valueStr
        ..longitude = valueStr
        ..maid = valueStr
        ..appVer = valueStr
        ..appVerNum = valueStr
        ..prevAppVer = valueStr
        ..prevAppVerNum = valueStr
        ..time = valueStr;

      return data;
    });

    await _eventRepo.insertAll(list);

    await _clearData();

    await sharedPreference.setBool(con.realmReserved, true);
  }

  _initRequest() {
    rabbitRequest
      ..sdkVersion = _versionController.sdkVersion
      ..globalId = _deviceInfoController.globalId
      ..osType = _deviceInfoController.osType
      ..osVer = _deviceInfoController.osVersion
      ..display = _deviceInfoController.display
      ..maid = _deviceInfoController.maId
      ..device = _deviceInfoController.device
      ..model = _deviceInfoController.model
      ..brandName = _deviceInfoController.brandName
      ..board = _deviceInfoController.board;
  }

  Future<void> checkOpenApp() async {
    SharedPreferences sharedPreference = await SharedPreferences.getInstance();
    // check run app first time if session start -> set session end
    if (sharedPreference.containsKey('sessionId') &&
        sharedPreference.containsKey('pauseTime')) {
      var pauseTime = sharedPreference.getString('pauseTime')!;
      _sessionId =
          sharedPreference.getString('sessionId') ?? Uuid.v4().toString();
      session = RabbitSession.start;
      Future.delayed(const Duration(seconds: 1), () async {
        await sessionEnd(
            endTime: DateTime.parse(pauseTime).format(con.dateFormat));
      });
    }

    await _eventRepo.deleteAfterRetentionDays(_config.retentionDays);
  }

  Future<void> rotationLoop() async {
    if (_config.collectionType == DataCollectionType.rotation.name) {
      if (!isLoading) await uploadEvent();
      rotationTask?.cancel();
      rotationTask =
          PausableTimer(Duration(minutes: _config.rotationTime), () async {
        //print('rotationLoop : ${DateTime.now()}');
        await uploadEvent();
        rotationTask!
          ..reset()
          ..start();
      });
      rotationTask!.start();
    }
  }

  Future<void> sessionStart() async {
    session = RabbitSession.start;
    _sessionId = Uuid.v4().toString();
    _sessionStart = DateTime.now();
    sessionTask?.cancel();
    sessionTask =
        PausableTimer(Duration(minutes: _config.sessionExpire), () async {
      await _loopTime();
    });
    sessionTask!.start();
    await addEvent(
        eventName: 'session_start',
        eventValue: '',
        dateTime: _sessionStart.format(con.dateFormat));
  }

  Future<void> _loopTime() async {
    if (session == RabbitSession.start) {
      _sessionId = Uuid.v4().toString();
      _sessionStart = DateTime.now();
      sessionTask!
        ..reset()
        ..start();
      await addEvent(
          eventName: 'session_start',
          eventValue: '',
          dateTime: _sessionStart.format(con.dateFormat));
    }
  }

  Future<void> sessionEnd({String? endTime}) async {
    if (session != RabbitSession.start) return;
    sessionTask?.cancel();
    session = RabbitSession.end;
    await addEvent(eventName: 'session_end', eventValue: '', dateTime: endTime);
    session = RabbitSession.empty;
    _sessionId = '';
    SharedPreferences sharedPreference = await SharedPreferences.getInstance();
    await sharedPreference.remove('pauseTime');
    await sharedPreference.remove('sessionId');
    //await _exportJson();
  }

  Future<void> onPause() async {
    SharedPreferences sharedPreference = await SharedPreferences.getInstance();
    await sharedPreference.setString(
        'pauseTime', DateTime.now().toIso8601String());
    await sharedPreference.setString('sessionId', _sessionId);
    sessionTask?.pause();
    resumeTask?.cancel();
  }

  Future<void> onResume() async {
    SharedPreferences sharedPreference = await SharedPreferences.getInstance();
    var pauseTime = sharedPreference.getString('pauseTime')!;
    var now = DateTime.now();
    var expireMilliseconds = _config.sessionAutoExpire * 60000;

    var inactiveTime = now.difference(DateTime.parse(pauseTime));

    if (inactiveTime.inMilliseconds > expireMilliseconds) {
      _sessionId =
          sharedPreference.getString('sessionId') ?? Uuid.v4().toString();

      await sessionEnd(
          endTime: _sessionStart
              .add(Duration(minutes: _config.sessionAutoExpire))
              .format(con.dateFormat));
      await sessionStart();
    } else {
      if (sessionTask != null && sessionTask!.isPaused) {
        var runTime = DateTime.parse(pauseTime).difference(_sessionStart);
        var sessionExpireMilliseconds = _config.sessionExpire * 60000;

        var leftTimeMilliseconds = sessionExpireMilliseconds -
            (runTime.inMilliseconds + inactiveTime.inMilliseconds);

        if (leftTimeMilliseconds < 0) {
          await sessionStart();
          return;
        }

        resumeTask = PausableTimer(Duration(milliseconds: leftTimeMilliseconds),
            () async {
          await _loopTime();
        });
        resumeTask!.start();
      }
    }

    await _eventRepo.deleteAfterRetentionDays(_config.retentionDays);
  }

  Future<void> _clearData() async {
    await _eventRepo.deleteAll();
    changeDB.add('deleteAll');
  }

  Future<void> addEvent(
      {required String eventName,
      required String eventValue,
      Map<dynamic, dynamic>? dataLayer,
      String? dateTime}) async {
    // refresh value
    await Future.wait<void>([
      _userInfoController.getData(),
      _locationController.getPosition(),
      _locationController.getLocale()
    ]);

    if (GetPlatform.isIOS &&
        _deviceInfoController.maId.isEmpty &&
        await Permission.appTrackingTransparency.isGranted) {
      await _deviceInfoController.getAdvertising();
    }

    var data = rabbitRequest
      ..memberId = _userInfoController.memberId
      ..ssoId = _userInfoController.ssoId
      ..sessionId = _sessionId
      ..sessionEvent = session
      ..sdkVersion = _versionController.sdkVersion
      ..eventName = eventName
      ..eventValue = eventValue
      ..osVer = _deviceInfoController.osVersion
      ..localeCountry = _locationController.localeCountry
      ..localeLang = _locationController.localeLang
      ..latitude = _locationController.position == null
          ? null
          : _checkData(
              _locationController.isGranted,
              _userInfoController.memberId,
              _config.consent,
              _locationController.position!.latitude.toString(),
              '00.00')
      ..longitude = _locationController.position == null
          ? null
          : _checkData(
              _locationController.isGranted,
              _userInfoController.memberId,
              _config.consent,
              _locationController.position!.longitude.toString(),
              '00.00')
      ..maid = _checkData(
          GetPlatform.isAndroid
              ? _config.permissionMaid
              : await Permission.appTrackingTransparency.isGranted,
          _userInfoController.memberId,
          _config.consent,
          _deviceInfoController.maId.toString(),
          '00000000-0000-0000-0000-000000000000')
      ..appVer = _versionController.appVer
      ..appVerNum = _versionController.appVerNum
      ..prevAppVer = _versionController.prevAppVer
      ..prevAppVerNum = _versionController.prevAppVerNum
      ..dataLayer = dataLayer
      ..time = dateTime ?? DateTime.now().format(con.dateFormat);

    if (_config.dbType == DBType.realm.name) {
      var last = await _eventRepo.last();
      if (last != null) {
        data.id = last.id + 1;
      }
    }

    await _conditionController.getStorageSize();
    // if storage < 1mb send to api
    if (_conditionController.currentFreeSpace < _config.storageMin) {
      var requests = <RabbitRequest>[];
      requests.add(data);
      await callAPI(requests);
      apiResult.add('storage = min');
      return;
    }
    // if storage between 1mb to 5mb delete oldest row
    else if (_conditionController.currentFreeSpace > _config.storageMin &&
        _conditionController.currentFreeSpace < _config.storageMax) {
      var data = await _eventRepo.first();
      if (data != null) {
        await _eventRepo.delete(data.id);
        apiResult.add('storage = deleteOld');
      }
    }

    // if free storage > 5mb insert normaly
    await _eventRepo.insert(data);
    changeDB.add('insert');

    // check collectionType = event : upload immediately
    if (_config.collectionType != DataCollectionType.event.name) {
      return;
    }

    if (!isLoading) await uploadEvent();
  }

  Future<List<RabbitRequest>> getEvent() async {
    return await _eventRepo.getAll();
  }

  Future<void> uploadEvent() async {
    // check offline mode not send
    if (_config.offlineMode) {
      apiResult.add('offlineMode = true');
      return;
    }

    var requests = await _eventRepo.getRequests();
    if (requests.isEmpty) return;

    // check cpu, memory, battery
    var validate = await _conditionController.checkSend();
    if (validate.containsValue(false)) {
      apiResult.add(
          'error [${validate.entries.where((element) => !element.value).map((e) => e.key).join(', ')}]');
      return;
    }

    await callAPI(requests);
  }

  Future<void> callAPI(List<RabbitRequest> data) async {
    // send data to api
    try {
      isLoading = true;
      final r = RetryOptions(
          delayFactor: const Duration(seconds: 10),
          maxAttempts: _config.retryTime);

      final response = await r.retry(() async {
        return await _api.post(
            con.eventBatchURL, data.map((e) => e.toJson()).toList());
      }, retryIf: (e) => e is SocketException || e is TimeoutException);

      if (response) {
        apiResult.add('uploadEvent success');

        await _eventRepo.updateSend(data.map((e) => e.id).toList());

        // check deleteAfterSent
        if (_config.deleteAfterSent) {
          await _eventRepo.deleteAllKeys(data.map((e) => e.id).toList());
        }

        changeDB.add('refresh');

        isLoading = false;

        // check collectionType = event
        if (_config.collectionType != DataCollectionType.event.name) return;

        // loop self method
        await uploadEvent();
      } else {
        apiResult.add('uploadEvent fail');
      }
    } on DioError catch (e) {
      apiResult.add('uploadEvent fail : ${e.message.toString()}');
    } finally {
      isLoading = false;
    }
  }

  Future<void> clearCache() async {
    await _eventRepo.deleteAllSent();
    changeDB.add('deleteAllSent');
  }

  Future<void> clearAll() async {
    await _eventRepo.deleteAll();
    changeDB.add('deleteAll');
  }

  // _exportJson() async {
  //   try {
  //     var results = await _eventRepo.getAll();
  //     var appDocDir = await getApplicationDocumentsDirectory();

  //     final File file = File('${appDocDir.path}/rabbit_data.json');
  //     await file.writeAsString(json.encode(results));
  //   } catch (e) {
  //     //print(e.toString());
  //   }
  // }

  String _checkData(bool isPermission, String memberId, bool? isConsent,
      dynamic value, String defaultValue) {
    if (!isPermission) return '';

    if (isConsent == null) {
      // User: Anonymous
      if (memberId.isEmpty) {
        return value;
      }
      // User: Member
      else {
        return defaultValue;
      }
    }

    if (!isConsent) {
      return defaultValue;
    }

    return value;
  }
}
